"""
Generate gridded lat/lon coordinates that can be used for fixed radius searches across a region.

Provides both an importable method and a CLI.

.. code:: console

    $ python app.py -h

- Jamie Taylor <jamie.taylor@sheffield.ac.uk>
- First Authored: 2021-11-16
"""
import argparse
import logging
import os
import sys
from typing import Optional

from pvoutput.grid_search.grid_search import GridSearch
from pvoutput.grid_search.natural_earth import NaturalEarth


def parse_options():
    """Parse command line options."""
    parser = argparse.ArgumentParser(
        description=("This is a command line interface (CLI) for " "the grid_search module."),
        epilog="Jamie Taylor, 2021-11-16",
    )
    parser.add_argument(
        "--bbox",
        metavar="<min_lat,min_lon,max_lat,max_lon>",
        dest="bbox",
        action="store",
        type=str,
        required=False,
        default=None,
        help="Specify a bounding box to search. Can be used in conjunction with " "--countries",
    )
    parser.add_argument(
        "--countries",
        metavar="<country>[,<country>[,...]]",
        dest="countries",
        action="store",
        type=str,
        required=False,
        default=None,
        help="Specify a list of countries, searching only grid points that fall "
        "within these countries' boundaries. Specify one or more countries, "
        "separated by commas (default is all). Country names must match those "
        "used in the Natural Earth dataset (HINT: run this code with the "
        "--list-countries option to list them). This option can be used in "
        "conjunction with --bbox, in which case the search will only include "
        "grid points within both the bounding box and the countries list.",
    )
    parser.add_argument(
        "--radial-clip",
        metavar="<lat,lon,radius_km>",
        dest="radial_clip",
        action="store",
        type=str,
        required=False,
        default=None,
        help="Specify a radius to clip to. Can be used in conjunction with --bbox "
        "and --countries. Pass the latitude, longitude and radius as a "
        "comma-separated string. Radius should be in km.",
    )
    parser.add_argument(
        "--list-countries",
        dest="list_countries",
        action="store_true",
        required=False,
        help="List the country names that can be used for the " "--countries option.",
    )
    parser.add_argument(
        "--buffer",
        metavar="<buffer>",
        dest="buffer",
        action="store",
        type=float,
        required=False,
        default=0.0,
        help="Specify a buffer/tolerance for including grid points i.e. include "
        "grid points that fall within <buffer> kilometers of the target "
        "boundary. Default is 0km.",
    )
    parser.add_argument(
        "--search-radius",
        metavar="<radius>",
        dest="search_radius",
        action="store",
        type=float,
        required=False,
        default=25.0,
        help="Specify the radial search limit around each grid point in " "kilometers.",
    )
    parser.add_argument(
        "--local-crs-epsg",
        metavar="<EPSG>",
        dest="local_crs_epsg",
        action="store",
        type=int,
        required=False,
        default=4087,
        help="Optionally provide the EPSG code of a local co-ordinate Reference "
        "System (CRS) for improved accuracy. e.g. set to 27700 (OSGB36 / "
        "British National Grid) if searching the British Isles.",
    )
    parser.add_argument(
        "--cache-dir",
        metavar="</path/to/dir>",
        dest="cache_dir",
        action="store",
        type=str,
        required=False,
        default=None,
        help="Specify a directory to use for caching downloaded boundary files.",
    )
    parser.add_argument(
        "--show",
        dest="show",
        action="store_true",
        required=False,
        help="Set this flag to show a plot of the grid.",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        metavar="<path>",
        dest="outfile",
        action="store",
        type=str,
        required=False,
        help="Specify a filename to save the grid to.",
    )
    options = parser.parse_args()

    def handle_options(options):
        """Validate command line args and pre-process where necessary."""
        if options.bbox is not None:
            options.bbox = list(map(lambda x: float(x.strip()), options.bbox.split(",")))
        if options.radial_clip is not None:
            options.radial_clip = list(
                map(lambda x: float(x.strip()), options.radial_clip.split(","))
            )
        if options.cache_dir is not None:
            if not os.path.isdir(options.cache_dir):
                logging.error(f"The cache_dir '{options.cache_dir}' does not exist.")
                raise ValueError(f"The cache_dir '{options.cache_dir}' does not exist.")
        if options.countries:
            options.countries = list(map(lambda x: str(x.strip()), options.countries.split(",")))
            earth = NaturalEarth(options.cache_dir)
            _, countries = earth.get_hires_world_boundaries()
            for country in options.countries:
                if country not in countries:
                    logging.error(f"The country '{country}' is invalid.")
                    raise ValueError(f"The country '{country}' is invalid.")
        if options.outfile is not None and os.path.exists(options.outfile):
            check = query_yes_no(
                f"The output file '{options.outfile}' already exists, results "
                "will be overwritten, do you want to continue?",
                "no",
            )
            if check is False:
                print("Quitting...")
                sys.exit(0)
        return options

    return handle_options(options)


def main():
    """Run main app"""
    options = parse_options()
    grd = GridSearch(cache_dir=options.cache_dir)
    if options.list_countries:
        grd.nat_earth.list_countries()
        sys.exit()
    grd.generate_grid(
        bbox=options.bbox,
        countries=options.countries,
        radial_clip=options.radial_clip,
        buffer=options.buffer,
        search_radius=options.search_radius,
        local_crs_epsg=options.local_crs_epsg,
        save_to=options.outfile,
        show=options.show,
    )


if __name__ == "__main__":
    fmt = "%(asctime)s [%(levelname)s] - %(message)s (%(filename)s:%(funcName)s)"
    datefmt = "%Y-%m-%dT%H:%M:%SZ"
    logging.basicConfig(format=fmt, datefmt=datefmt, level=os.environ.get("LOGLEVEL", "INFO"))
    main()


def query_yes_no(question: str, default: Optional[str] = "yes") -> bool:
    """Ask a yes/no question via input() and return the answer as boolean.

    Args:
        question:
            The question presented to the user.
        default:
            The presumed answer if the user just hits <Enter>. It must be "yes" (the default), "no"
            or None (meaning an answer is required of the user).

    Returns:
        Return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)
    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")



"""
Grid Search Class used to make grid of latitude and longitude coordinates
"""

from typing import Iterable, Optional

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pyproj import Transformer

from pvoutput.grid_search.clip import (
    bounding_box_from_radius,
    buffer_bounding_box_bounds,
    clip_to_bbox,
    clip_to_countries,
    clip_to_radius,
)
from pvoutput.grid_search.natural_earth import NaturalEarth


class GridSearch:
    """A class for generating a gridded search."""

    def __init__(self, cache_dir: str = None) -> None:
        """Initialise.

        Args:
            cache_dir: Optionally provide a location to cache boundary
            definition files locally and avoid unnecsessary downloads.
        """
        self.nat_earth = NaturalEarth(cache_dir)

    def plot_grid(
        self,
        coords: gpd.GeoDataFrame,
        countries: Iterable[str],
        bbox: Optional[Iterable[float]] = None,
        local_crs_epsg: int = 4087,
        filename: Optional[str] = None,
    ) -> None:
        """Plot grid coordinates.

        Plot grid coordinates over world boundaries with selected countries highlighted.

        Args:
            coords: A geopandas GeoDataFrame containing latitudes, longitudes and geometries
                for a set of coordinates.
            countries: A list of country names to clip the coords to.
            bbox: Optionally pass a four element iterable defining a bounding box:
                [min_lat, min_lon, max_lat, max_lon]. This will be used to set the scale of the
                plot.
            local_crs_epsg: Optionally provide the EPSG code of a local
                Co-ordinate Reference System (CRS) for improved accuracy.
                e.g. set to 27700 (OSGB36 / British National Grid) if searching
                the British Isles. The default is EPSG:4087 (a.k.a WGS 84 / World Equidistant
                Cylindrical), which works globally but with less accuracy.
            filename: Optionally pass a filename (relative or absolute) to save the plot to.
                Image format should be set using the file extension (i.e. .jpeg, .png or .svg).
        """
        world, _ = self.nat_earth.get_hires_world_boundaries()
        world.to_crs(f"EPSG:{local_crs_epsg}", inplace=True)
        coords.to_crs(f"EPSG:{local_crs_epsg}", inplace=True)
        if bbox is None:
            bbox = [
                coords.latitude.min(),
                coords.longitude.min(),
                coords.latitude.max(),
                coords.longitude.max(),
            ]
        f, ax = plt.subplots()
        world.plot(
            ax=ax, color="palegreen", edgecolor="black", linewidth=1, label="World", zorder=1
        )
        if countries is not None:
            selected = world[world.name.isin(countries)]
            selected.geometry.boundary.plot(
                ax=ax, color=None, edgecolor="gold", label="Selected countries", zorder=2
            )
        coords.plot(ax=ax, marker="o", color="red", markersize=2, label="Grid", zorder=3)
        xmin = coords.geometry.bounds.minx.min()
        xmax = coords.geometry.bounds.maxx.max()
        ymin = coords.geometry.bounds.miny.min()
        ymax = coords.geometry.bounds.maxy.max()
        xpadding = (xmax - xmin) / 8
        ypadding = (ymax - ymin) / 8
        ax.set_xlim(xmin - xpadding, xmax + xpadding)
        ax.set_ylim(ymin - ypadding, ymax + ypadding)
        ax.axes.xaxis.set_visible(False)
        ax.axes.yaxis.set_visible(False)
        plt.legend(prop={"size": 6})
        plt.show()
        if filename is not None:
            plt.savefig(filename, dpi=300)

    def generate_grid(
        self,
        bbox: Optional[Iterable[float]] = None,
        countries: Optional[Iterable[str]] = None,
        radial_clip: Optional[Iterable[float]] = None,
        buffer: float = 0,
        search_radius: float = 25,
        local_crs_epsg: int = 4087,
        save_to: Optional[str] = None,
        show: bool = False,
    ):
        """Use hexagonal tiling to generate a grid search with minimal overlap.

        Create a set of gridded coordinates which use hexagonal tiling as an efficient way to
        conduct a fixed-radius search of a region defined by a bounding box and/or country borders.

        Args:
            bbox:
                Optionally pass a four element iterable defining a bounding box:
                [min_lat, min_lon, max_lat, max_lon].
            countries:
                Optionally pass a list of country names to clip the coords to.
            radial_clip:
                Optionally set a radial boundary to clip to. Pass a three element iterable
                containing: (<lat>, <lon>, <radius_km>).
            buffer:
                Optionally buffer the bounding box and country boundaries before clipping, in
                kilometers.
            search_radius:
                Optionally set the radial search limit around each grid point in kilometers.
                Defaults to 25km.
            local_crs_epsg:
                Optionally provide the EPSG code of a local co-ordinate Reference System (CRS) for
                improved accuracy. e.g. set to 27700 (OSGB36 / British National Grid) if searching
                the British Isles. The default is EPSG:4087 (a.k.a WGS 84 / World Equidistant
                Cylindrical), which works globally but with poor accuracy in some locations.
            save_to:
                Optionally specify a filename to save the grid co-ordinates to (CSV).
            show:
                Set to True to show a plot of the grid.

        Returns:
            A pandas DataFrame containing co-ordinates for the grid with columns: latitude,
            longitude.
        """
        # get countries
        world, all_countries = self.nat_earth.get_hires_world_boundaries()
        countries = all_countries if countries is None else countries

        # create bounding box
        if bbox is None:
            if radial_clip is None:
                bbox = (
                    world[world.name.isin(countries)]
                    .dissolve()
                    .buffer(buffer * 1000.0)
                    .to_crs("EPSG:4326")
                    .bounds.to_numpy()[0]
                )
                bbox = [bbox[1], bbox[0], bbox[3], bbox[2]]
            else:
                bbox = bounding_box_from_radius(
                    radial_clip[0], radial_clip[1], radial_clip[2], local_crs_epsg
                )
        bbox = buffer_bounding_box_bounds(bbox, buffer)
        bounds = [np.round(b, 5) for b in bbox]

        # create x and y bounds
        search_radius_m = search_radius * 1000.0
        wgs84_to_projected = Transformer.from_crs(4326, local_crs_epsg, always_xy=True)
        projected_to_wgs84 = Transformer.from_crs(local_crs_epsg, 4326, always_xy=True)
        xmin, ymin = wgs84_to_projected.transform(bounds[1], bounds[0])
        xmax, ymax = wgs84_to_projected.transform(bounds[3], bounds[2])
        y_interval = search_radius_m * np.cos(np.radians(30))
        x_interval = search_radius_m * 3
        x_offset = 0

        # create coordinates
        coords = []
        for y in np.arange(ymin - y_interval * 3, ymax + y_interval + search_radius_m, y_interval):
            xmin_ = xmin - search_radius_m - x_offset
            xmax_ = xmax + x_interval + search_radius_m + x_offset
            for x in np.arange(xmin_, xmax_, x_interval):
                coords.append(projected_to_wgs84.transform(x, y))
            if x_offset == 0:
                x_offset = search_radius_m * 1.5
            else:
                x_offset = 0
        coords = pd.DataFrame(coords, columns=["longitude", "latitude"])
        coords = gpd.GeoDataFrame(
            coords, geometry=gpd.points_from_xy(coords.longitude, coords.latitude)
        )
        coords = coords.set_crs("EPSG:4326").to_crs("EPSG:4087")
        coords = clip_to_countries(
            coords=coords,
            world=world,
            countries=countries,
            buffer=buffer,
            search_radius=search_radius,
        )
        coords = clip_to_bbox(
            coords=coords,
            bbox=bbox,
            buffer=buffer,
            search_radius=search_radius,
            local_crs_epsg=local_crs_epsg,
        )
        if radial_clip is not None:
            coords = clip_to_radius(
                coords=coords,
                latitude=radial_clip[0],
                longitude=radial_clip[1],
                radius=radial_clip[2],
                search_radius=search_radius,
                local_crs_epsg=local_crs_epsg,
            )

        # show plot
        if show:
            self.plot_grid(
                coords,
                countries,
                bbox,
            )

        # save plots
        if save_to is not None:
            coords.to_csv(
                save_to, float_format="%.5f", index=False, columns=["longitude", "latitude"]
            )
        return coords[["latitude", "longitude"]].reset_index(drop=True)



import os
from datetime import date

import numpy as np
import pandas as pd
import pytest

from pvoutput import utils
from pvoutput.daterange import DateRange

PV_SYSTEM = 123


def test_get_missing_dates_for_id(data_dir):
    test_hdf = os.path.join(data_dir, "data/test.hdf")
    missing_dates = utils.get_missing_dates_for_id(test_hdf, PV_SYSTEM)
    np.testing.assert_array_equal(missing_dates, [date(2019, 1, 2), date(2019, 1, 3)])


def test_get_system_ids_in_store(data_dir):
    test_hdf = os.path.join(data_dir, "data/test.hdf")
    system_ids = utils.get_system_ids_in_store(test_hdf)
    np.testing.assert_array_equal(system_ids, [PV_SYSTEM])


def test_get_date_ranges_to_download(data_dir):
    test_hdf = os.path.join(data_dir, "data/test.hdf")
    date_ranges = utils.get_date_ranges_to_download(test_hdf, PV_SYSTEM, "2018-01-01", "2019-01-10")
    # 2018-01-02 and 2018-01-03 are already known to be missing.
    np.testing.assert_array_equal(
        date_ranges,
        [
            DateRange(start_date=date(2018, 1, 1), end_date=date(2018, 12, 31)),
            DateRange(start_date=date(2019, 1, 4), end_date=date(2019, 1, 10)),
        ],
    )


def test_datetime_list_to_dates():
    np.testing.assert_array_equal(
        utils.datetime_list_to_dates(pd.Timestamp("2019-01-01")), [date(2019, 1, 1)]
    )

    np.testing.assert_array_equal(
        utils.datetime_list_to_dates([pd.Timestamp("2019-01-01"), pd.Timestamp("2019-01-02")]),
        [date(2019, 1, 1), date(2019, 1, 2)],
    )



"""Date Range Class """
from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Iterable, List, Union

import numpy as np
import pandas as pd


@dataclass
class DateRange:
    """Date Range Class"""

    start_date: date
    end_date: date

    def __init__(self, start_date, end_date):
        """Init"""
        self.start_date = safe_convert_to_date(start_date)
        self.end_date = safe_convert_to_date(end_date)

    def intersection(self, other):
        """
        Get intersection of this DateRange and other
        """
        new_start = max(self.start_date, other.start_date)
        new_end = min(self.end_date, other.end_date)
        if new_start < new_end:
            return DateRange(new_start, new_end)

    def date_range(self) -> np.array:
        """
        Make date range
        """
        return pd.date_range(self.start_date, self.end_date, freq="D").date

    def total_days(self) -> int:
        """
        Find the total number of days
        """
        return (
            np.timedelta64(self.end_date - self.start_date)
            .astype("timedelta64[D]")
            .astype(np.float32)
        )

    def split_into_years(self) -> List:
        """
        Split start and end dates into list of years

        """
        duration = self.end_date - self.start_date
        num_years = duration / timedelta(days=365)
        if num_years <= 1:
            return [self]
        else:
            end_date = self.end_date
            new_date_ranges = []
            for year_back in range(np.ceil(num_years).astype(int)):
                start_date = end_date - timedelta(days=365)
                if start_date < self.start_date:
                    start_date = self.start_date
                new_date_ranges.append(DateRange(start_date, end_date))
                end_date = start_date
            return new_date_ranges


def get_date_range_list(dates: Iterable[date]) -> List[DateRange]:
    """
    Get data range lists for dates

    Args:
        dates: list of dates

    Returns: list of date ranges
    """
    if not dates:
        return []
    dates = np.array(dates)
    dates = np.sort(dates)
    dates_diff = np.diff(dates)
    location_of_gaps = np.where(dates_diff > timedelta(days=1))[0]
    index_of_last_date = len(dates) - 1
    location_of_gaps = list(location_of_gaps)
    location_of_gaps.append(index_of_last_date)

    start_i = 0
    date_range_list = []
    for end_i in location_of_gaps:
        date_range = DateRange(start_date=dates[start_i], end_date=dates[end_i])
        date_range_list.append(date_range)
        start_i = end_i + 1

    return date_range_list


def safe_convert_to_date(dt: Union[datetime, date, str]) -> date:
    """
    Convert datetime to date

    Args:
        dt: datetime, date or string

    Returns: date
    """
    if isinstance(dt, str):
        dt = pd.Timestamp(dt)
    if isinstance(dt, datetime):
        return dt.date()
    if isinstance(dt, date):
        return dt


def merge_date_ranges_to_years(date_ranges: Iterable[DateRange]) -> List[DateRange]:
    """
    Merge date ranges to years

    Args:
        date_ranges: List of DateRanges, in ascending chronological order.

    Returns:
        List of DateRanges, each representing a year, in descending order.
    """
    if not date_ranges:
        return []

    # Split multi-year date ranges
    date_ranges_split = []
    for date_range in date_ranges[::-1]:
        date_ranges_split.extend(date_range.split_into_years())

    years_to_download = []
    for date_range in date_ranges_split:
        if years_to_download:
            intersection = date_range.intersection(years_to_download[-1])
            if intersection == date_range:
                # date_range falls within the last year to retrieve,
                # so we can ignore this date_range
                continue
            elif intersection is None:
                # No overlap
                date_to = date_range.end_date
            else:
                # Overlap
                date_to = intersection.start_date

        else:
            date_to = date_range.end_date

        date_from = date_to - timedelta(days=365)
        years_to_download.append(DateRange(date_from, date_to))

    return years_to_download



""" Function to process data """
import logging
from io import StringIO

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def process_system_status(pv_system_status_text, date) -> pd.DataFrame:
    """
    Process raw system status

    Args:
        pv_system_status_text: string of system data, like:
            "1234;07:45,21,255,1,2;07:50,21,255,1;07:50,21,255,1,2"
        date: The date this data is from

    Returns: dataframe of data
    """

    # See https://pvoutput.org/help/data_services.html#data-services-get-system-status
    columns = [
        "cumulative_energy_gen_Wh",
        "instantaneous_power_gen_W",
        "temperature_C",
        "voltage",
    ]
    if pv_system_status_text == "no status found":
        logger.debug("Text was empty so return empty dataframe")
        return pd.DataFrame(columns=columns + ["system_id", "datetime"])

    # get system id
    system_id = int(pv_system_status_text.split(";")[0])
    pv_system_status_text = ";".join(pv_system_status_text.split(";")[1:])

    try:
        one_pv_system_status = pd.read_csv(
            StringIO(pv_system_status_text),
            lineterminator=";",
            names=["time"] + columns,
            dtype={col: np.float64 for col in columns},
        ).sort_index()

    except Exception as e:

        # this can happen if there is only one data value and it doesnt contain all 5 columns.
        # if there is many rows of data, then it seems fine
        if pv_system_status_text.count(";") != 0:
            # the data contains more than one row, so lets raise the error
            raise e

        # how many columns does it have
        n_columns = pv_system_status_text.count(",") + 1

        one_pv_system_status = pd.read_csv(
            StringIO(pv_system_status_text),
            lineterminator=";",
            names=["time"] + columns[: n_columns - 1],
            dtype={col: np.float64 for col in columns},
        ).sort_index()

        missing_columns = [c for c in columns if c not in one_pv_system_status.columns]
        one_pv_system_status[missing_columns] = np.NAN

    # process dataframe
    one_pv_system_status["system_id"] = system_id

    # format date
    one_pv_system_status["date"] = date
    one_pv_system_status["date"] = pd.to_datetime(date)

    # format time
    one_pv_system_status["time"] = pd.to_datetime(one_pv_system_status["time"]).dt.strftime(
        "%H:%M:%S"
    )
    one_pv_system_status["time"] = pd.to_timedelta(one_pv_system_status["time"])

    # make datetime
    one_pv_system_status["datetime"] = one_pv_system_status["date"] + one_pv_system_status["time"]
    one_pv_system_status.drop(columns=["date", "time"], inplace=True)

    return one_pv_system_status


def process_batch_status(pv_system_status_text) -> pd.DataFrame:
    """
    Process batch status text

    Args:
        pv_system_status_text: text to be procssed

    Returns: dataframe of data

    """
    # See https://pvoutput.org/help.html#dataservice-getbatchstatus

    # PVOutput uses a non-standard format for the data.  The text
    # needs some processing before it can be read as a CSV.
    processed_lines = []
    for line in pv_system_status_text.split("\n"):
        line_sections = line.split(";")
        date = line_sections[0]
        time_and_data = line_sections[1:]
        processed_line = [
            "{date},{payload}".format(date=date, payload=payload) for payload in time_and_data
        ]
        processed_lines.extend(processed_line)

    if processed_lines:
        first_line = processed_lines[0]
        num_cols = len(first_line.split(","))
        if num_cols >= 8:
            raise NotImplementedError("Handling of consumption data is not implemented!")

    processed_text = "\n".join(processed_lines)
    del processed_lines

    columns = ["cumulative_energy_gen_Wh", "instantaneous_power_gen_W", "temperature_C", "voltage"]

    pv_system_status = pd.read_csv(
        StringIO(processed_text),
        names=["date", "time"] + columns,
        parse_dates={"datetime": ["date", "time"]},
        index_col=["datetime"],
        dtype={col: np.float64 for col in columns},
    ).sort_index()

    return pv_system_status



""" Util functions"""
import logging
import os
import sys
import warnings
from datetime import date, datetime
from typing import Dict, Iterable, List, Union

import numpy as np
import pandas as pd
import requests
import tables
import yaml
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from pvoutput.consts import CONFIG_FILENAME
from pvoutput.daterange import DateRange, get_date_range_list

_LOG = logging.getLogger("pvoutput")


def _get_param_from_config_file(param_name, config_filename=CONFIG_FILENAME):
    with open(config_filename, mode="r") as fh:
        config_data = yaml.load(fh, Loader=yaml.Loader)
    try:
        value = config_data[param_name]
    except KeyError as e:
        print("Config file", config_filename, "does not contain a", param_name, "parameter.", e)
        raise
    return value


def get_logger(filename=None, mode="a", level=logging.DEBUG, stream_handler=False):
    """
    Get a logger

    Args:
        filename: get file handler filename
        mode: file handler mode
        level: logging level
        stream_handler: option to make a stream handler aswell

    Returns: logger
    """
    if filename is None:
        filename = _get_param_from_config_file("log_filename")
    logger = logging.getLogger("pvoutput")
    logger.setLevel(level)
    logger.handlers = [logging.FileHandler(filename=filename, mode=mode)]
    if stream_handler:
        logger.handlers.append(logging.StreamHandler(sys.stdout))
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    for handler in logger.handlers:
        handler.setFormatter(formatter)

    # Attach urllib3's logger to our logger.
    loggers_to_attach = ["urllib3", "requests"]
    for logger_name_to_attach in loggers_to_attach:
        logger_to_attach = logging.getLogger(logger_name_to_attach)
        logger_to_attach.parent = logger
        logger_to_attach.propagate = True

    return logger


def _get_session_with_retry() -> requests.Session:
    max_retry_counts = dict(
        connect=720,  # How many connection-related errors to retry on.
        # Set high because sometimes the network goes down for a
        # few hours at a time.
        # 720 x Retry.MAX_BACKOFF (120 s) = 86,400 s = 24 hrs
        read=20,  # How many times to retry on read errors.
        status=20,  # How many times to retry on bad status codes.
    )
    retries = Retry(
        total=max(max_retry_counts.values()),
        backoff_factor=0.5,
        status_forcelist=[500, 502, 503, 504],
        **max_retry_counts
    )
    session = requests.Session()
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))
    return session


def _get_response(api_url: str, api_params: Dict, headers: Dict) -> requests.Response:
    api_params_str = "&".join(["{}={}".format(key, value) for key, value in api_params.items()])
    full_api_url = "{}?{}".format(api_url, api_params_str)
    session = _get_session_with_retry()
    response = session.get(full_api_url, headers=headers)
    _LOG.debug("response: status_code=%d; headers=%s", response.status_code, response.headers)
    return response


def _print_and_log(msg: str, level: int = logging.INFO):
    _LOG.log(level, msg)
    print(msg)


def get_system_ids_in_store(store_filename: str) -> List[int]:
    """
    Get system ids in the hdf store

    Args:
        store_filename: hdf file name

    Returns: list of systems ids
    """
    if not os.path.exists(store_filename):
        return []
    with pd.HDFStore(store_filename, mode="r") as store:
        pv_system_ids = list(store.walk("/timeseries"))[0][2]
    return pd.to_numeric(pv_system_ids)


def get_date_ranges_to_download(
    store_filename: str,
    system_id: int,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
) -> List[DateRange]:
    """
    Get the date ranges that we need downloaded

    If system_id in store, check if it already has data from
    start_date to end_date, taking into consideration missing_dates.

    Returns: list of DateRange objects
        For each DateRange we need to download from
        start_date to end_date inclusive.
    """
    dates_to_download = list(pd.date_range(start_date, end_date, freq="D"))
    dates_to_download = datetime_list_to_dates(dates_to_download)
    dates_already_downloaded = get_dates_already_downloaded(store_filename, system_id)
    dates_to_download = set(dates_to_download) - set(dates_already_downloaded)
    missing_dates_for_id = get_missing_dates_for_id(store_filename, system_id)
    dates_to_download -= set(missing_dates_for_id)
    return get_date_range_list(list(dates_to_download))


def get_missing_dates_for_id(store_filename: str, system_id: int) -> List:
    """
    Get missing dates for on pv system id

    Args:
        store_filename: filename fo hdf store
        system_id: system id

    Returns: list of missing dates
    """
    if not os.path.exists(store_filename):
        return []

    with pd.HDFStore(store_filename, mode="r") as store:
        missing_dates_for_id = store.select(
            key="missing_dates",
            where="index=system_id",
            columns=["missing_start_date_PV_localtime", "missing_end_date_PV_localtime"],
        )

    missing_dates = []
    for _, row in missing_dates_for_id.iterrows():
        missing_date_range = pd.date_range(
            row["missing_start_date_PV_localtime"], row["missing_end_date_PV_localtime"], freq="D"
        ).tolist()
        missing_dates.extend(missing_date_range)

    missing_dates = np.sort(np.unique(missing_dates))
    missing_dates = datetime_list_to_dates(missing_dates)
    print()
    _LOG.info("system_id %d: %d missing dates already found", system_id, len(missing_dates))
    return missing_dates


def datetime_list_to_dates(datetimes: Iterable[datetime]) -> Iterable[date]:
    """
    Change datetime list to dates

    Args:
        datetimes: list of datetimes

    Returns: datetime index of dates
    """
    if not isinstance(datetimes, Iterable):
        datetimes = [datetimes]
    return pd.DatetimeIndex(datetimes).date


def get_dates_already_downloaded(store_filename, system_id) -> set:
    """
    Get the dates that have already been downloaded

    Args:
        store_filename: filename of hdf file
        system_id: one system id

    Returns: set of datetimes already downloaded

    """
    if not os.path.exists(store_filename):
        return set([])

    with pd.HDFStore(store_filename, mode="r") as store:
        key = system_id_to_hdf_key(system_id)
        try:
            datetimes = store.select(key=key, columns=["datetime", "query_date"])
        except KeyError:
            return set([])
        else:
            query_dates = datetime_list_to_dates(datetimes["query_date"].dropna())
            return set(datetimes.index.date).union(query_dates)


def system_id_to_hdf_key(system_id: int) -> str:
    """
    Change system id to a hdf key

    Args:
        system_id: system id

    Returns: key
    """
    return "/timeseries/{:d}".format(system_id)


def sort_and_de_dupe_pv_system(store, pv_system_id):
    """
    Sort and de-duplicate pv systems

    Args:
        store: store of pv systems
        pv_system_id: on pv system id

    """
    key = system_id_to_hdf_key(pv_system_id)
    timeseries = store[key]
    timeseries.sort_index(inplace=True)
    timeseries = timeseries[~timeseries.index.duplicated()]
    store.remove(key)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", tables.NaturalNameWarning)
        store.append(key, timeseries, data_columns=True)



""" Constants used in this repo """
import os
from datetime import timedelta
from urllib.parse import urljoin

BASE_URL = "https://pvoutput.org"
MAP_URL = urljoin(BASE_URL, "map.jsp")
REGIONS_URL = urljoin(BASE_URL, "region.jsp")

# Country codes used by PVOutput.org on, for example,
# https://pvoutput.org/map.jsp.  Taken from
# https://pvoutput.org/help.html#api-addsystem.

PV_OUTPUT_COUNTRY_CODES = {
    "New South Wales": 1,
    "Afghanistan": 2,
    "Akrotiri": 3,
    "Albania": 4,
    "Algeria": 5,
    "American Samoa": 6,
    "Andorra": 7,
    "Angola": 8,
    "Anguilla": 9,
    "Antarctica": 10,
    "Antigua and Barbuda": 11,
    "Arctic Ocean": 12,
    "Argentina": 13,
    "Armenia": 14,
    "Aruba": 15,
    "Ashmore and Cartier Islands": 16,
    "Atlantic Ocean": 17,
    "Austria": 18,
    "Azerbaijan": 19,
    "Bahamas, The": 20,
    "Bahrain": 21,
    "Bangladesh": 22,
    "Barbados": 23,
    "Belarus": 24,
    "Belgium": 25,
    "Belize": 26,
    "Benin": 27,
    "Bermuda": 28,
    "Bhutan": 29,
    "Bolivia": 30,
    "Bosnia and Herzegovina": 31,
    "Botswana": 32,
    "Bouvet Island": 33,
    "Brazil": 34,
    "British Indian Ocean Territory": 35,
    "British Virgin Islands": 36,
    "Brunei": 37,
    "Bulgaria": 38,
    "Burkina Faso": 39,
    "Burma": 40,
    "Burundi": 41,
    "Cambodia": 42,
    "Cameroon": 43,
    "Canada": 44,
    "Cape Verde": 45,
    "Cayman Islands": 46,
    "Central African Republic": 47,
    "Chad": 48,
    "Chile": 49,
    "China": 50,
    "Christmas Island": 51,
    "Clipperton Island": 52,
    "Cocos (Keeling) Islands": 53,
    "Colombia": 54,
    "Comoros": 55,
    "Congo, Democratic Republic of the": 56,
    "Congo, Republic of the": 57,
    "Cook Islands": 58,
    "Coral Sea Islands": 59,
    "Costa Rica": 60,
    "Cote d'Ivoire": 61,
    "Croatia": 62,
    "Cuba": 63,
    "Curacao": 64,
    "Cyprus": 65,
    "Czech Republic": 66,
    "Denmark": 67,
    "Dhekelia": 68,
    "Djibouti": 69,
    "Dominica": 70,
    "Dominican Republic": 71,
    "Ecuador": 72,
    "Egypt": 73,
    "El Salvador": 74,
    "Equatorial Guinea": 75,
    "Eritrea": 76,
    "Estonia": 77,
    "Ethiopia": 78,
    "Falkland Islands": 79,
    "Faroe Islands": 80,
    "Fiji": 81,
    "Finland": 82,
    "France": 83,
    "French Polynesia": 84,
    "French Southern and Antarctic Lands": 85,
    "Gabon": 86,
    "Gambia, The": 87,
    "Gaza Strip": 88,
    "Georgia": 89,
    "Germany": 90,
    "Ghana": 91,
    "Gibraltar": 92,
    "Greece": 93,
    "Greenland": 94,
    "Grenada": 95,
    "Guam": 96,
    "Guatemala": 97,
    "Guernsey": 98,
    "Guinea": 99,
    "Guinea-Bissau": 100,
    "Guyana": 101,
    "Haiti": 102,
    "Heard Island and McDonald Islands": 103,
    "Holy See (Vatican City)": 104,
    "Honduras": 105,
    "Hong Kong": 106,
    "Hungary": 107,
    "Iceland": 108,
    "India": 109,
    "Indian Ocean": 110,
    "Indonesia": 111,
    "Iran": 112,
    "Iraq": 113,
    "Ireland": 114,
    "Isle of Man": 115,
    "Israel": 116,
    "Italy": 117,
    "Jamaica": 118,
    "Jan Mayen": 119,
    "Japan": 120,
    "Jersey": 121,
    "Jordan": 122,
    "Kazakhstan": 123,
    "Kenya": 124,
    "Kiribati": 125,
    "Korea, North": 126,
    "Korea, South": 127,
    "Kosovo": 128,
    "Kuwait": 129,
    "Kyrgyzstan": 130,
    "Laos": 131,
    "Latvia": 132,
    "Lebanon": 133,
    "Lesotho": 134,
    "Liberia": 135,
    "Libya": 136,
    "Liechtenstein": 137,
    "Lithuania": 138,
    "Luxembourg": 139,
    "Macau": 140,
    "Macedonia": 141,
    "Madagascar": 142,
    "Malawi": 143,
    "Malaysia": 144,
    "Maldives": 145,
    "Mali": 146,
    "Malta": 147,
    "Marshall Islands": 148,
    "Mauritania": 149,
    "Mauritius": 150,
    "Mayotte": 151,
    "Mexico": 152,
    "Micronesia": 153,
    "Moldova": 154,
    "Monaco": 155,
    "Mongolia": 156,
    "Montenegro": 157,
    "Montserrat": 158,
    "Morocco": 159,
    "Mozambique": 160,
    "Namibia": 161,
    "Nauru": 162,
    "Navassa Island": 163,
    "Nepal": 164,
    "Netherlands": 165,
    "New Caledonia": 166,
    "New Zealand": 167,
    "Nicaragua": 168,
    "Niger": 169,
    "Nigeria": 170,
    "Niue": 171,
    "Norfolk Island": 172,
    "Northern Mariana Islands": 173,
    "Norway": 174,
    "Oman": 175,
    "Pakistan": 176,
    "Palau": 177,
    "Panama": 178,
    "Papua New Guinea": 179,
    "Paracel Islands": 180,
    "Paraguay": 181,
    "Peru": 182,
    "Philippines": 183,
    "Pitcairn Islands": 184,
    "Poland": 185,
    "Portugal": 186,
    "Puerto Rico": 187,
    "Qatar": 188,
    "Romania": 189,
    "Russia": 190,
    "Rwanda": 191,
    "Saint Barthelemy": 192,
    "Saint Helena, Ascension, and Tristan da Cunha": 193,
    "Saint Kitts and Nevis": 194,
    "Saint Lucia": 195,
    "Saint Martin": 196,
    "Saint Pierre and Miquelon": 197,
    "Saint Vincent and the Grenadines": 198,
    "Samoa": 199,
    "San Marino": 200,
    "Sao Tome and Principe": 201,
    "Saudi Arabia": 202,
    "Senegal": 203,
    "Serbia": 204,
    "Seychelles": 205,
    "Sierra Leone": 206,
    "Singapore": 207,
    "Sint Maarten": 208,
    "Slovakia": 209,
    "Slovenia": 210,
    "Solomon Islands": 211,
    "Somalia": 212,
    "South Africa": 213,
    "South Georgia and South Sandwich Is.": 214,
    "Southern Ocean": 215,
    "Spain": 216,
    "Spratly Islands": 217,
    "Sri Lanka": 218,
    "Sudan": 219,
    "Suriname": 220,
    "Svalbard": 221,
    "Swaziland": 222,
    "Sweden": 223,
    "Switzerland": 224,
    "Syria": 225,
    "Taiwan": 226,
    "Tajikistan": 227,
    "Tanzania": 228,
    "Thailand": 229,
    "Timor-Leste": 230,
    "Togo": 231,
    "Tokelau": 232,
    "Tonga": 233,
    "Trinidad and Tobago": 234,
    "Tunisia": 235,
    "Turkey": 236,
    "Turkmenistan": 237,
    "Turks and Caicos Islands": 238,
    "Tuvalu": 239,
    "Uganda": 240,
    "Ukraine": 241,
    "United Arab Emirates": 242,
    "United Kingdom": 243,
    "United States": 244,
    "Uruguay": 245,
    "Uzbekistan": 246,
    "Vanuatu": 247,
    "Venezuela": 248,
    "Vietnam": 249,
    "Virgin Islands": 250,
    "Wake Island": 251,
    "Wallis and Futuna": 252,
    "West Bank": 253,
    "Western Sahara": 254,
    "Yemen": 255,
    "Zambia": 256,
    "Zimbabwe": 257,
}

PV_OUTPUT_MAP_COLUMN_NAMES = {
    "timeseries_duration": "c",
    "average_generation_per_day": "avg",
    "efficiency": "gss",
    "power_generation": "atg",
    "capacity": "ss",
    "address": "o",
    "name": "sn",
}


ONE_DAY = timedelta(days=1)

PV_OUTPUT_DATE_FORMAT = "%Y%m%d"
CONFIG_FILENAME = os.environ.get("PVOUTPUT_CONFIG", os.path.expanduser("~/.pvoutput.yml"))
RATE_LIMIT_PARAMS_TO_API_HEADERS = {
    "rate_limit_remaining": "X-Rate-Limit-Remaining",
    "rate_limit_total": "X-Rate-Limit-Limit",
    "rate_limit_reset_time": "X-Rate-Limit-Reset",
}



from datetime import date

import pandas as pd

from pvoutput import daterange
from pvoutput.daterange import DateRange, merge_date_ranges_to_years


def test_get_date_range_list():
    def _get_date_range(start_date, periods):
        return list(pd.date_range(start_date, periods=periods, freq="D"))

    dates = []
    for start_date, periods in [("2019-01-01", 5), ("2019-05-01", 3), ("2015-04-01", 1)]:
        dates.extend(_get_date_range(start_date, periods))

    date_range_list = daterange.get_date_range_list(dates)
    assert date_range_list[0].start_date == date(2015, 4, 1)
    assert date_range_list[0].end_date == date(2015, 4, 1)

    assert date_range_list[1].start_date == date(2019, 1, 1)
    assert date_range_list[1].end_date == date(2019, 1, 5)

    assert date_range_list[2].start_date == date(2019, 5, 1)
    assert date_range_list[2].end_date == date(2019, 5, 3)

    assert daterange.get_date_range_list([]) == []


def test_intersection():
    assert (
        DateRange("2019-01-01", "2019-01-02").intersection(DateRange("2020-01-01", "2020-01-02"))
        is None
    )

    assert DateRange("2019-01-01", "2019-01-10").intersection(
        DateRange("2019-01-01", "2019-01-02")
    ) == DateRange("2019-01-01", "2019-01-02")

    assert DateRange("2019-01-01", "2019-01-10").intersection(
        DateRange("2019-01-05", "2019-01-20")
    ) == DateRange("2019-01-05", "2019-01-10")

    year = DateRange("2018-01-1", "2019-01-01")
    dec = DateRange("2018-12-01", "2019-01-01")
    assert year.intersection(dec) == dec

    june = DateRange("2018-06-01", "2018-07-01")
    assert year.intersection(june) == june

    incomplete_overlap = DateRange("2017-07-01", "2018-02-01")
    assert year.intersection(incomplete_overlap) != incomplete_overlap


def test_total_days():
    assert DateRange("2019-01-01", "2019-01-10").total_days() == 9


def test_split_into_years():
    short_dr = DateRange("2019-01-01", "2019-01-10")
    assert short_dr.split_into_years() == [short_dr]

    one_year = DateRange("2019-01-01", "2020-01-01")
    assert one_year.split_into_years() == [one_year]

    year_and_half = DateRange("2019-01-01", "2020-06-01")
    assert year_and_half.split_into_years() == [
        DateRange("2019-06-02", "2020-06-01"),
        DateRange("2019-01-01", "2019-06-02"),
    ]


def test_merge_date_ranges_to_years():
    jan = DateRange("2018-01-01", "2018-02-01")
    multiyear = DateRange("2017-01-01", "2018-02-01")
    old_multiyear = DateRange("2014-01-01", "2016-02-01")
    ancient_jan = DateRange("2010-01-01", "2010-02-01")
    for date_ranges, merged in [
        ([], []),
        ([jan], [DateRange("2017-02-01", "2018-02-01")]),
        (
            [multiyear],
            [DateRange("2017-02-01", "2018-02-01"), DateRange("2016-02-02", "2017-02-01")],
        ),
        (
            [old_multiyear, multiyear],
            [
                DateRange("2017-02-01", "2018-02-01"),
                DateRange("2016-02-02", "2017-02-01"),
                DateRange("2015-02-01", "2016-02-01"),
                DateRange("2014-02-01", "2015-02-01"),
                DateRange("2013-02-01", "2014-02-01"),
            ],
        ),
        (
            [ancient_jan, old_multiyear, multiyear],
            [
                DateRange("2017-02-01", "2018-02-01"),
                DateRange("2016-02-02", "2017-02-01"),
                DateRange("2015-02-01", "2016-02-01"),
                DateRange("2014-02-01", "2015-02-01"),
                DateRange("2013-02-01", "2014-02-01"),
                DateRange("2009-02-01", "2010-02-01"),
            ],
        ),
    ]:
        assert merge_date_ranges_to_years(date_ranges) == merged



""" Custom Exeception classes """
import requests


class BadStatusCode(Exception):
    """Bad status code excepction"""

    def __init__(self, response: requests.Response, message: str = ""):
        """Init"""
        self.response = response
        super(BadStatusCode, self).__init__(message)

    def __str__(self) -> str:
        """String method"""
        string = super(BadStatusCode, self).__str__()
        string += "Status code: {}; ".format(self.response.status_code)
        string += "Response content: {}; ".format(self.response.content)
        string += "Response headers: {}; ".format(self.response.headers)
        return string


class NoStatusFound(BadStatusCode):
    """Exeception for when no status code is found"""

    pass


class RateLimitExceeded(BadStatusCode):
    """Class for rate limit is exceeded"""

    pass
